/*
 * Copyright (C) 2015 ~ 2017 Deepin Technology Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package main

import log "github.com/cihub/seelog"
import "pkg.deepin.io/lib/gettext"
import "dbus/org/freedesktop/notifications"
import "fmt"

type Action struct {
	Id       string
	Name     string
	Callback func()
}

func SendNotify(icon string, msg string, actions []Action) {
	log.Infof("Notify: %q, %v\n", msg, actions)
	n, err := notifications.NewNotifier("org.freedesktop.Notifications", "/org/freedesktop/Notifications")

	var as []string
	for _, action := range actions {
		as = append(as, action.Id, action.Name)
	}

	if icon == "" {
		icon = "deepin-appstore"
	}
	id, err := n.Notify("appstore", 0, icon, "", msg, as, nil, -1)
	if err != nil {
		log.Warnf("Notify failed: %q: %v\n", msg, err)
		return
	}

	n.ConnectNotificationClosed(func(_id uint32, reason uint32) {
		if id == id {
			notifications.DestroyNotifier(n)
		}
	})
	n.ConnectActionInvoked(func(_id uint32, actionId string) {
		if id != _id {
			return
		}
		for _, action := range actions {
			if action.Id == actionId && action.Callback != nil {
				action.Callback()
			}
		}
	})
}

// NotifyInstall send desktop notify for install job
func NotifyInstall(pkgId string, succeed bool, ac []Action) {
	var msg string
	if succeed {
		msg = fmt.Sprintf(gettext.Tr("%q installed successfully."), pkgId)
		SendNotify("package_install_succeed", msg, ac)
	} else {
		msg = fmt.Sprintf(gettext.Tr("%q failed to install."), pkgId)
		SendNotify("package_install_failed", msg, ac)
	}

}

func NotifyRemove(pkgId string, succeed bool, ac []Action) {
	var msg string
	if succeed {
		msg = fmt.Sprintf(gettext.Tr("%q removed successfully."), pkgId)
	} else {
		msg = fmt.Sprintf(gettext.Tr("%q failed to remove."), pkgId)
	}
	SendNotify("deepin-appstore", msg, ac)
}

// NotifyDownload send desktop notify for download job
func NotifyFailedDownload(pkgName string, ac []Action) {
	msg := fmt.Sprintf(gettext.Tr("%q failed to download."), pkgName)
	SendNotify("package_download_failed", msg, ac)
}

//NotifyLowPower send notify for low power
func NotifyLowPower() {
	msg := gettext.Tr("In order to prevent automatic shutdown, please plug in for normal update.")
	SendNotify("notification-battery_low", msg, nil)
}

func NotifyUpgrade(succeed bool, ac []Action) {
	var msg string
	if succeed {
		if FileExist("/var/run/reboot-required.pkgs") {
			msg = gettext.Tr("Updated successfully!") + " " + gettext.Tr("Some functions will take effect after rebooting.")
		} else {
			msg = gettext.Tr("Updated successfully!")
		}
		SendNotify("package_install_succeed", msg, ac)
	} else {
		msg = gettext.Tr("Failed to update.")
		SendNotify("package_update_failed", msg, ac)
	}

}

func (l *Lastore) notifyDownloadUpgradablePackages(nApps int, hasLibs bool) {
	var msg string
	switch {
	case nApps > 0 && !hasLibs:
		msg = fmt.Sprintf(gettext.Tr("%d software need to be updated."), nApps)
	case nApps == 0 && hasLibs:
		msg = fmt.Sprintf(gettext.Tr("Some patches need to be updated."))
	case nApps > 0 && hasLibs:
		msg = fmt.Sprintf(gettext.Tr("Some patches and %d software need to be updated."), nApps)
	default:
		return
	}

	SendNotify("system_updated",
		msg, []Action{Action{
			Id:   "download",
			Name: gettext.Tr("Download"),
			Callback: func() {
				l.core.PrepareDistUpgrade()
			},
		}})
}

func notifyDownloadUpgradablePackagesFailed(ac []Action) {
	msg := gettext.Tr("Failed to download upgradeable packages.")
	SendNotify("package_download_failed", msg, ac)
}

func notifyAutoClean() {
	msg := gettext.Tr("Package cache wiped")
	SendNotify("deepin-appstore", msg, nil)
}

func notifySourceModified(actions []Action) {
	msg := gettext.Tr("Your system source has been modified, please restore to official source for your normal use")
	SendNotify("dialog-warning", msg, actions)
}
